package org.xpresso.xml;

import java.io.File;
import java.io.FileWriter;
import java.io.Reader;
import java.io.StringReader;
import java.util.ArrayDeque;
import java.util.Deque;
import java.util.Map;
import java.util.Set;

import org.xpresso.xml.output.Outputter;

/**
 *
 * The main class, used to generate the XML document. It contains a stack containing {@link org.xpresso.xml.Element Elements}.<br />
 * Content is added to the upper Element on the stack. Document provides methods to stack and unstack elements,<br />
 * adding {@link org.xpresso.xml.XmlContent content} and returning the resulting XML.<br />
 * <br />
 * Document is itself a XmlContent, so a Document can be added at a chosen place of another Document. Header information (XML header,<br />
 * DTD declaration) are in this case discarded.
 * <br />
 * This code is under the <a href="http://www.gnu.org/licenses/lgpl.html">LGPL v3 licence</a>.
 * @author Alexis Dufrenoy
 *
 */
public class Document extends XmlContent implements Cloneable, CharSequence {
	
	/**
	 * 
	 */
	private static final long serialVersionUID = 7863288457178265868L;

	/**
	 * Contains the header: XML declaration and DTD reference
	 */
	protected StringBuilder header;

	/**
	 * This stack is used to know the depth in the document tree
	 */
	protected Deque<Element> stack;

	/**
	 * Content can only be added to the Document if this boolean is true. It is set to false by<br />
	 * the {@link org.xpresso.xml.Document#finish() finish()} method or when<br />
	 * the last {@link org.xpresso.xml.Element Element} is unstacked by<br />
	 * the {@link org.xpresso.xml.Document#unstackElement() unstackElement()} method.
	 */
	protected boolean open;

	/**
	 * Constructor creating a new Document with only a root {@link org.xpresso.xml.Element Element}.
	 * @param root Root {@link org.xpresso.xml.Element Element} of the document
	 */
	protected Document(Element root) {
		header = new StringBuilder("");
		open=true;
		stack = new ArrayDeque<Element>();
		content = new StringBuilder();
		try {
			stackElement(root);
		} catch (ClosedDocumentException cde) {
			//cannot happen
		}
	}
	
	/**
	 * Creates a new Document with only a root {@link org.xpresso.xml.Element Element}.
	 * @param root Root {@link org.xpresso.xml.Element Element} of the document
	 * @return The created Document
	 */
	public static Document createDocument(Element root) {
		return new Document(root);
	}

	/**
	 * Constructor creating a new Document with a xml version and a root {@link org.xpresso.xml.Element Element}.
	 * @param xmlVersion XML version in the header of the Document
	 * @param root Root {@link org.xpresso.xml.Element Element} of the document
	 */
	protected Document(String xmlVersion, Element root) {
		open=true;
		stack = new ArrayDeque<Element>();
		content = new StringBuilder();
		header = new StringBuilder();
		header.append("<?xml version=\"")
			.append(xmlVersion)
			.append("\"?>");
		try {
			stackElement(root);
		} catch (ClosedDocumentException cde) {
			//cannot happen
		}
	}
	
	/**
	 * Creates a new Document with a xml version and a root {@link org.xpresso.xml.Element Element}.
	 * @param xmlVersion XML version in the header of the Document
	 * @param root Root {@link org.xpresso.xml.Element Element} of the document
	 * @return The created Document
	 */
	public static Document createDocument(String xmlVersion, Element root) {
		return new Document(xmlVersion,root);
	}

	/**
	 * Constructor creating a new Document with a xml version, a root {@link org.xpresso.xml.Element Element}<br />
	 * and a DOCTYPE statement.
	 * @param xmlVersion XML version in the header of the Document
	 * @param root Root {@link org.xpresso.xml.Element Element} of the document
	 * @param dtdName Reference name to give to the DTD with the XML document
	 * @param dtdFile DTD File
	 */
	protected Document(String xmlVersion, Element root, String dtdName, String dtdFile) {
		open=true;
		stack = new ArrayDeque<Element>();
		content = new StringBuilder();
		header = new StringBuilder();
		header.append("<?xml version=\"")
			.append(xmlVersion)
			.append("\"?>")
			.append("<!DOCTYPE ")
			.append(dtdName)
			.append(" SYSTEM \"")
			.append(dtdFile)
			.append("\">");
		try {
			stackElement(root);
		} catch (ClosedDocumentException cde) {
			//cannot happen
		}
	}
	
	/**
	 * Creates a new Document with a xml version, a root {@link org.xpresso.xml.Element Element}<br />
	 * and a DOCTYPE statement.
	 * @param xmlVersion XML version in the header of the Document
	 * @param root Root {@link org.xpresso.xml.Element Element} of the document
	 * @param dtdName Reference name to give to the DTD with the XML document
	 * @param dtdFile DTD File
	 * @return The created Document
	 * 
	 */
	public static Document createDocument(String xmlVersion, Element root, String dtdName, String dtdFile) {
		return new Document(xmlVersion,root,dtdName,dtdFile);
	}

	/**
	 * Constructor creating a new Document with a xml version, the character set encoding<br />
	 * and a root {@link org.xpresso.xml.Element Element}.
	 * @param xmlVersion XML version in the header of the Document
	 * @param encoding Character set encoding
	 * @param root Root {@link org.xpresso.xml.Element Element} of the document
	 */
	protected Document (String xmlVersion, String encoding, Element root) {
		open=true;
		stack = new ArrayDeque<Element>();
		content = new StringBuilder();
		header = new StringBuilder();
		header.append("<?xml version=\"")
			.append(xmlVersion)
			.append("\" encoding=\"")
			.append(encoding)
			.append("\"?>");
		try {
			stackElement(root);
		} catch (ClosedDocumentException cde) {
			//cannot happen
		}
	}
	
	/**
	 * Creates a new Document with a xml version, the character set encoding<br />
	 * and a root {@link org.xpresso.xml.Element Element}.
	 * @param xmlVersion XML version in the header of the Document
	 * @param encoding Character set encoding
	 * @param root Root {@link org.xpresso.xml.Element Element} of the document
	 * @return The created Document
	 */
	public static Document createDocument(String xmlVersion, String encoding, Element root) {
		return new Document(xmlVersion,encoding,root);
	}

	/**
	 * Constructor creating a new Document with a xml version, the character set encoding,<br />
	 * a root {@link org.xpresso.xml.Element Element} and a DOCTYPE statement.
	 * @param xmlVersion XML version in the header of the Document
	 * @param encoding Character set encoding
	 * @param root Root {@link org.xpresso.xml.Element Element} of the document
	 * @param dtdName Reference name to give to the DTD with the XML document
	 * @param dtdFile DTD File
	 */
	protected Document (String xmlVersion, String encoding, Element root, String dtdName, String dtdFile) {
		open=true;
		stack = new ArrayDeque<Element>();
		content = new StringBuilder();
		header = new StringBuilder();
		header.append("<?xml version=\"")
			.append(xmlVersion)
			.append("\" encoding=\"")
			.append(encoding)
			.append("\"?>")
			.append("<!DOCTYPE ")
			.append(dtdName)
			.append(" SYSTEM \"")
			.append(dtdFile)
			.append("\">");
		try {
			stackElement(root);
		} catch (ClosedDocumentException cde) {
			//cannot happen
		}
	}
	
	/**
	 * a new Document with a xml version, the character set encoding,<br />
	 * a root {@link org.xpresso.xml.Element Element} and a DOCTYPE statement.
	 * @param xmlVersion XML version in the header of the Document
	 * @param encoding Character set encoding
	 * @param root Root {@link org.xpresso.xml.Element Element} of the document
	 * @param dtdName Reference name to give to the DTD with the XML document
	 * @param dtdFile DTD File
	 * @return The created Document
	 */
	public static Document createDocument(String xmlVersion, String encoding, Element root, String dtdName, String dtdFile) {
		return new Document(xmlVersion,encoding,root,dtdName,dtdFile);
	}

	/**
	 * Constructor creating a new Document with a xml version, the character set encoding, the standalone setting<br />
	 * and a root {@link org.xpresso.xml.Element Element}.
	 * @param xmlVersion XML version in the header of the Document
	 * @param encoding Character set encoding
	 * @param root Root {@link org.xpresso.xml.Element Element} of the document
	 * @param standalone Is the document standalone, which means does it refer to a DTD or not. Values are "true" or "false"
	 */
	protected Document (String xmlVersion, String encoding, Element root, boolean standalone) {
		open=true;
		stack = new ArrayDeque<Element>();
		content = new StringBuilder();
		String st=(standalone?"yes":"no");
		header = new StringBuilder();
		header.append("<?xml version=\"")
			.append(xmlVersion)
			.append("\" encoding=\"")
			.append(encoding)
			.append("\" standalone=\"")
			.append(st)
			.append("\"?>");
		try {
			stackElement(root);
		} catch (ClosedDocumentException cde) {
			//cannot happen
		}
	}
	
	/**
	 * Creates a new Document with a xml version, the character set encoding, the standalone setting<br />
	 * and a root {@link org.xpresso.xml.Element Element}.
	 * @param xmlVersion XML version in the header of the Document
	 * @param encoding Character set encoding
	 * @param root Root {@link org.xpresso.xml.Element Element} of the document
	 * @param standalone Is the document standalone, which means does it refer to a DTD or not. Values are "true" or "false"
	 * @return The created Document
	 */
	public static Document createDocument(String xmlVersion, String encoding, Element root, boolean standalone) {
		return new Document(xmlVersion,encoding,root,standalone);
	}

	/**
	 * Constructor creating a new Document with a xml version, the character set encoding, the standalone setting<br />
	 * a root {@link org.xpresso.xml.Element Element} and a DOCTYPE statement.
	 * @param xmlVersion XML version in the header of the Document
	 * @param encoding Character set encoding
	 * @param root Root {@link org.xpresso.xml.Element Element} of the document
	 * @param standalone Is the document standalone, which means does it refer to a DTD or not. Values are "true" or "false"
	 * @param dtdName Reference name to give to the DTD with the XML document
	 * @param dtdFile DTD File
	 */
	protected Document (String xmlVersion, String encoding, Element root, boolean standalone, String dtdName, String dtdFile) {
		open=true;
		stack = new ArrayDeque<Element>();
		content = new StringBuilder();
		String st=(standalone?"yes":"no");
		header = new StringBuilder();
		header.append("<?xml version=\"")
			.append(xmlVersion)
			.append("\" encoding=\"")
			.append(encoding)
			.append("\" standalone=\"")
			.append(st)
			.append("\"?>")
			.append("<!DOCTYPE ")
			.append(dtdName)
			.append(" SYSTEM \"")
			.append(dtdFile)
			.append("\">");
		try {
			stackElement(root);
		} catch (ClosedDocumentException cde) {
			//cannot happen
		}
	}
	
	/**
	 * Creates a new Document with a xml version, the character set encoding, the standalone setting<br />
	 * a root {@link org.xpresso.xml.Element Element} and a DOCTYPE statement.
	 * @param xmlVersion XML version in the header of the Document
	 * @param encoding Character set encoding
	 * @param root Root {@link org.xpresso.xml.Element Element} of the document
	 * @param standalone Is the document standalone, which means does it refer to a DTD or not. Values are "true" or "false"
	 * @param dtdName Reference name to give to the DTD with the XML document
	 * @param dtdFile DTD File
	 * @return The created Document
	 */
	public static Document createDocument(String xmlVersion, String encoding, Element root, boolean standalone, String dtdName, String dtdFile) {
		return new Document(xmlVersion,encoding,root,standalone,dtdName,dtdFile);
	}

	/**
	 * Constructor creating a new Document without content. All data structures are initialized and empty
	 */
	protected Document() {
		header = new StringBuilder("");
		stack = new ArrayDeque<Element>();
		content = new StringBuilder();
	}

	/**
	 * Adds a level of depth to the current branch of the document. It adds an {@link org.xpresso.xml.Element Element}.<br />
	 * Further content will be added to the newly added element, until a new one is stacked or it is unstacked.
	 * @param element The new Element to be stacked
	 * @throws ClosedDocumentException Thrown if the method is added after the document was closed
	 */
	public void stackElement(Element element) throws ClosedDocumentException {
		if (open) {
			content.append('<').append(element.getLabel());
			Map<String,String> attributes = element.getAttributes();
			Set<String> entries = attributes.keySet();
			for (String entry : entries) {
				content.append(' ').append(entry).append("=\"").append(attributes.get(entry)).append('"');
			}
			content.append('>');
			stack.addFirst(element);
		} else {
			throw new ClosedDocumentException("Impossible to stack element : document is closed");
		}
	}

	/**
	 * Unstack a level of depth from the current branch of the document. Further {@link org.xpresso.xml.XmlContent XmlContent}<br />
	 * will be added to the {@link org.xpresso.xml.Element Element} just below the one unstacked. If this one is the last in<br />
	 * the stack, the document is closed.
	 * @throws ClosedDocumentException Thrown if the method is added after the document was closed
	 */
	public void unstackElement() throws ClosedDocumentException {
		if (open) {
			content.append("</")
				.append(stack.pollFirst().getLabel())
				.append('>');
			//stack.removeFirst();
			if (stack.isEmpty()) {
				open=false;
			}
		} else {
			throw new ClosedDocumentException("Impossible to unstack element : document is closed");
		}
	}

	/**
	 * Add an {@link org.xpresso.xml.XmlContent XmlContent} to the uppest {@link org.xpresso.xml.Element Element} of the stack.<br />
	 * If the content is an Element, it will be added as an empty element.
	 * @param xmlContent Content to be added
	 * @throws ClosedDocumentException Thrown if the method is added after the document was closed
	 */
	public void add(XmlContent xmlContent) throws ClosedDocumentException {
		if (open) {
			content.append(xmlContent.getText());
		}
	}

	/**
	 * Unstack all {@link org.xpresso.xml.Element Elements}, which means the document is closed.
	 */
	public void finish() {
		while(open && !stack.isEmpty()) {
			try {
				unstackElement();
			} catch (ClosedDocumentException cde) {
				//cannot happen
			}
		}
	}

	/**
	 * Return the Document in its current state of completion. If {@link org.xpresso.xml.Element Elements} are left on the stack,<br />
	 * the result will not be a well-formed XML document.
	 * @return The document as a String
	 */
	public String getXmlText() {
		StringBuilder ret = new StringBuilder(header);
		ret.append(content);
		return ret.toString();
	}
	
	/**
	 * Returns a StringReader containing the XML.
	 * @return The XML
	 */
	public StringReader getStringReader() {
		StringBuilder ret = new StringBuilder(header);
		ret.append(content);
		return new StringReader(ret.toString());
	}
	
	/**
	 * Return the Document as a String which can be added to another Document, which means it has to be<br />
	 * closed and that the header informations are discarded.
	 * @return String containing the XmlContent
	 * @throws IllegalContentException Thrown if the Document is not closed
	 */
	@Override
	public String getText() throws IllegalContentException{
		if (this.open) {
			throw new IllegalContentException("Not closed Document cannot be added to another Document");
		}
		return content.toString();
	}
	
	/**
	 * Required for implementing CharSequence
	 */
	@Override
	public char charAt(int index) {
		return (getXmlText().charAt(index));
	}

	/**
	 * Required for implementing CharSequence
	 */
	@Override
	public int length() {
		return (getXmlText().length());
	}

	/**
	 * Required for implementing CharSequence
	 */
	@Override
	public CharSequence subSequence(int start, int end) {
		return (getXmlText().subSequence(start, end));
	}
	
	/**
	 * Returns a deep clone of the Document, regardless its current state. Can be used to generated several Documents<br />
	 * with approaching content
	 */
	@Override
	public Document clone() throws CloneNotSupportedException {
		Document ret = new Document();
		ret.header = new StringBuilder(this.header.toString());
		ret.open = this.open;
		ret.content = new StringBuilder(this.content.toString());
		for (Element elem : stack) {
			ret.stack.add(elem.clone());
		}
		return ret;
	}
	
	/**
	 * Creates a new Element
	 * @param label label of the new Element
	 * @return A new element
	 */
	public static Element createElement(String label) {
		return new Element(label); 
	}

	/**
	 * Returns a new CDATA
	 * @param text Content of the CDATA
	 * @return A new CDATA block
	 */
	public static CDATA createCdata(String text) {
		return new CDATA(text);
	}
	
	/**
	 * Returns a new Comment
	 * @param text Content of the Comment
	 * @return A new Comment
	 */
	public static Comment createComment(String text) {
		return new Comment(text);
	}
	
	/**
	 * Returns a new ProcessingInstruction
	 * @param text Text of the Processing Instruction
	 * @return a new Processing Instruction
	 */
	public static ProcessingInstruction createProcessingInstruction(String text) {
		return new ProcessingInstruction(text);
	}
	
	/**
	 * Returns a new XML Text
	 * @param text Content of the Text
	 * @return A new Text
	 */
	public static Text createText(String text) {
		return new Text(text);
	}

	/**
	 * For test purpose only !
	 * @param args
	 */
	public static void main(String args[]) {
		try {
			Document doc = Document.createDocument("1.0",new Element("toto"));
			Element elm=Document.createElement("titi");
			elm.addAttribute("id", "753");
			doc.stackElement(elm);
			doc.stackElement(new Element("tutu"));
			doc.unstackElement();
			doc.stackElement(new Element("tata"));
			
			Document doc2 = Document.createDocument(new Element("toto2"));
			doc2.add(Document.createElement("titi2"));
			doc2.finish();
			
			doc.add(doc2);
			
			doc.finish();
			
			FileWriter fw = new FileWriter(new File("d:\\toto.xml"));
			Outputter out = new Outputter();
			out.output(doc, fw);
			fw.flush();
			fw.close();
			/*System.out.println(doc.getXmlText());
			
			Document doc3 = doc.clone();
			System.out.println(doc3.getXmlText());*/
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

}
