package org.xpresso.xml;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 *
 * Element represents a level in the XML tree. If stacked in the {@link org.xpresso.xml.Document Document},<br />
 * any further {@link org.xpresso.xml.XmlContent content} will be added as child of this Element.<br />
 * <br />
 * This code is under the <a href="http://www.gnu.org/licenses/lgpl.html">LGPL v3 licence</a>.
 * @author Alexis Dufrenoy
 *
 */
public class Element extends XmlContent implements Cloneable {

	/**
	 * 
	 */
	private static final long serialVersionUID = -4208839732216090823L;

	/**
	 * Text of the element tag
	 */
	private final String label;

	/**
	 * Map to add attributes to the element
	 */
	private Map<String,String> attributes;

	/**
	 * Constructor creating a new Element using a given label.
	 * @param label The label of the Element
	 * @throws IllegalContentException Thrown if label contains invalid characters. Valid characters are uppercase and<br />
	 * lower case letters, digits, ":", "-", "_" and "."
	 */
	protected Element(String label) throws IllegalContentException {
		if (isValid(label)) {
			this.label=label;
			attributes = new HashMap<String,String>();
		} else {
			throw new IllegalContentException("The string "+label+" is illegal as a XML tag");
		}
	}

	/**
	 * Adds an attribute to the Element, using given key and value. 
	 * @param key the key part of the attribute
	 * @param value the value part of the attribute
	 * @throws IllegalContentException Thrown if label contains invalid characters. Valid characters are uppercase and<br />
	 * lower case letters, numbers, ":", "-", "_" and "."
	 */
	public void addAttribute(String key, String value) throws IllegalContentException {
		if (isValid(key)) {
			attributes.put(key, value);
		} else {
			throw new IllegalContentException("The string "+key+" is illegal as a XML attribute name");
		}
	}

	/**
	 * Returns the label of the Element
	 * @return the label of the attribute
	 */
	public String getLabel() {
		return label;
	}

	/**
	 * Return the map containing all attributes
	 * @return the map
	 */
	public Map<String, String> getAttributes() {
		return attributes;
	}

	/**
	 * Return the Element as an empty element
	 * @return text for an empty element
	 */
	public String getText() {
		StringBuilder text = new StringBuilder();
		text.append('<').append(getLabel());
		Set<String> entries = attributes.keySet();
		for (String entry : entries) {
			text.append(' ').append(entry).append("=\"").append(attributes.get(entry)).append('"');
		}
		text.append(" />");
		return text.toString();
	}

	/**
	 * Tests if the given String is a valid element or attribut name
	 * @param content text to be tested
	 * @return true if the text is valid, false else
	 */
    public static boolean isValid(String content) {
		boolean ret=true;
        if (content.matches("[^a-zA-Z0-9\\-_\\.\\:]+")) {
        	ret=false;
        }
        return ret;
    }
    
    @Override
	public Element clone() throws CloneNotSupportedException {
    	Element ret = new Element(this.label);
    	Set<String> keys = attributes.keySet();
    	Map<String,String> retAtt = new HashMap<String,String>();
    	for (String key : keys) {
    		retAtt.put(key, attributes.get(key));
    	}
    	ret.attributes=retAtt;
    	return ret;
    }

}
