package org.xpresso.xml.concurrent;

import java.util.ArrayDeque;
import java.util.Deque;
import java.util.Map;
import java.util.Set;

import org.xpresso.xml.ClosedDocumentException;
import org.xpresso.xml.Document;
import org.xpresso.xml.Element;
import org.xpresso.xml.IllegalContentException;
import org.xpresso.xml.XmlContent;

/**
 * Thread-safe version of Document. Methods are synchronized and the stack can only contain ConcurrentElement.<br />
 * <br />
 * This code is under the <a href="http://www.gnu.org/licenses/lgpl.html">LGPL v3 licence</a>.
 * @author Alexis Dufrenoy
 *
 */
public class ConcurrentDocument extends Document {
	
	/**
	 * 
	 */
	private static final long serialVersionUID = -5577235113700152378L;
	
	protected Deque<ConcurrentElement> stack = new ArrayDeque<ConcurrentElement>();;

	public ConcurrentDocument(ConcurrentElement root) {
		super(root);
	}

	public ConcurrentDocument(String xmlVersion, ConcurrentElement root, String dtdName, String dtdFile) {
		super(xmlVersion,root,dtdName,dtdFile);
	}

	public ConcurrentDocument(String xmlVersion, ConcurrentElement root) {
		super(xmlVersion,root);
	}

	public ConcurrentDocument(String xmlVersion, String encoding, ConcurrentElement root, boolean standalone, String dtdName, String dtdFile) {
		super(xmlVersion,encoding,root,standalone,dtdName,dtdFile);
	}

	public ConcurrentDocument(String xmlVersion, String encoding, ConcurrentElement root, boolean standalone) {
		super(xmlVersion,encoding,root,standalone);
	}

	public ConcurrentDocument(String xmlVersion, String encoding, ConcurrentElement root, String dtdName, String dtdFile) {
		super(xmlVersion,encoding,root,dtdName,dtdFile);
	}

	public ConcurrentDocument(String xmlVersion, String encoding, ConcurrentElement root) {
		super();
	}

	@Override
	public synchronized void add(XmlContent xmlContent) throws ClosedDocumentException {
		super.add(xmlContent);
	}

	@Override
	public synchronized void finish() {
		super.finish();
	}

	@Override
	public synchronized String getText() throws IllegalContentException {
		return super.getText();
	}

	@Override
	public synchronized String getXmlText() {
		return super.getXmlText();
	}

	@Override
	public synchronized void stackElement(Element element) throws ClosedDocumentException {
		if (!(element instanceof ConcurrentElement)) {
			throw new ClassCastException ("Only ConcurrentElements can be stacked in a ConcurrentDocument");
		} else {
			super.stackElement(element);
		}
	}

	@Override
	public synchronized void unstackElement() throws ClosedDocumentException {
		super.unstackElement();
	}
	
	public static Document createDocument(Element root) {
		return new ConcurrentDocument((ConcurrentElement)root);
	}
	
	public static Document createDocument(String xmlVersion, Element root) {
		return new ConcurrentDocument(xmlVersion,(ConcurrentElement)root);
	}
	
	public static Document createDocument(String xmlVersion, Element root, String dtdName, String dtdFile) {
		return new ConcurrentDocument(xmlVersion,(ConcurrentElement)root,dtdName,dtdFile);
	}
	
	public static Document createDocument(String xmlVersion, String encoding, Element root) {
		return new ConcurrentDocument(xmlVersion,encoding,(ConcurrentElement)root);
	}
	
	public static Document createDocument(String xmlVersion, String encoding, Element root, String dtdName, String dtdFile) {
		return new ConcurrentDocument(xmlVersion,encoding,(ConcurrentElement)root,dtdName,dtdFile);
	}
	
	public static Document createDocument(String xmlVersion, String encoding, Element root, boolean standalone) {
		return new ConcurrentDocument(xmlVersion,encoding,(ConcurrentElement)root,standalone);
	}
	
	public static Document createDocument(String xmlVersion, String encoding, Element root, boolean standalone, String dtdName, String dtdFile) {
		return new ConcurrentDocument(xmlVersion,encoding,(ConcurrentElement)root,standalone,dtdName,dtdFile);
	}
	
	public static Element createElement(String label) {
		return new ConcurrentElement(label); 
	}
	
	/**
	 * For test purpose only !
	 * @param args
	 */
	public final static void main(String args[]) {
		try {
			ConcurrentDocument doc = (ConcurrentDocument)ConcurrentDocument.createDocument("1.0",ConcurrentDocument.createElement("toto"));
			Element elm=ConcurrentDocument.createElement("titi");
			elm.addAttribute("id", "753");
			doc.stackElement(elm);
			doc.stackElement(ConcurrentDocument.createElement("tutu"));
			doc.unstackElement();
			doc.stackElement(ConcurrentDocument.createElement("tata"));
			
			ConcurrentDocument doc2 = (ConcurrentDocument)ConcurrentDocument.createDocument(ConcurrentDocument.createElement("toto2"));
			doc2.add(Document.createElement("titi2"));
			doc2.finish();
			
			doc.add(doc2);
			
			doc.finish();
			System.out.println(doc.getXmlText());
			
			Document doc3 = doc.clone();
			System.out.println(doc3.getXmlText());
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

}
